// INCLUDES
#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

// CONFIG
#pragma config FOSC = HS        // Oscillator Selection bits (HS oscillator: High-speed crystal/resonator on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config MCLRE = OFF      // RA5/MCLR/VPP Pin Function Select bit (RA5/MCLR/VPP pin function is digital input, MCLR internally tied to VDD)
#pragma config BOREN = OFF      // Brown-out Detect Enable bit (BOD disabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable bit (RB4/PGM pin has digital I/O function, HV on MCLR must be used for programming)
#pragma config CPD = OFF        // Data EE Memory Code Protection bit (Data memory code protection off)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit (Code protection off)

// DEFINITIONS
#define _XTAL_FREQ 16000000     // Reference freqw for the compiler
#define Rs PORTAbits.RA0        // LCD Rs connected to RA0
#define En PORTAbits.RA1        // LCD En connected to RA1
#define Fan PORTAbits.RA2       // Fan controller pin
#define Bat PORTAbits.RA3       // Charger controller pin

#define lcd_port TRISB          // LCD port is B
#define lcdline1 0x00           // Line 1
#define lcdline2 0x40           // Line 2

// GLOBAL VARIABLES
bool getchar_active = false;
unsigned char ch, i;
unsigned char gps_header[]="GPRMC,";
unsigned char gps_time[7];
unsigned char buffer[7];
unsigned char gps_fix[2];

// FUNCTION PROTOTYPE
int lcd_strobe(void);
void lcd_write(unsigned char dat);
void lcd_cmd(unsigned char cmd);
void lcd_data(unsigned char dat);
void lcd_init(void);
void lcd_writestr(const unsigned char *c);
void lcd_setcursor(unsigned char loc);
void lcd_line1(void);
void lcd_line2(void);
void showWelcome(void);
void uart_init(void);
unsigned char getch(void);
void interrupt ISR(void);
void gps_waitforfix(void);
void fan_on(void);


// FUNCTIONS
// LCD FUNCTIONS
// Function to strobe the enable pin
int lcd_strobe(void)
{
    En=1;
    __delay_us(1);
    En=0;
}

// Function to write to the lcd port
void lcd_write(unsigned char dat)
{
    lcd_port &= 0x0f;       // Get current port and clear upper bits
    lcd_port |= (dat & 0xf0);   // Combine upper and lower, leave lower
    lcd_strobe();           // Pulse enable
    
    lcd_port &= 0x0f;       
    lcd_port |= ((dat << 4) & (0xf0));
    lcd_strobe();
    __delay_ms(2);    
}

// Function to set LCDin command mode
void lcd_cmd(unsigned char cmd)
{
    Rs = 0;         // Rs low is LCD command mode
    lcd_write(cmd);
}

// Function to set the LCD in data mode 
void lcd_data(unsigned char dat)
{
    Rs = 1;
    lcd_write(dat);
}

// Function to initilize the LCD
void lcd_init(void)
{
    lcd_port &= 0x0f;   // Clear upper bits
    lcd_port |= 0x30;   // Send data to D7-D4
    Rs=0;
    lcd_strobe();
    __delay_ms(10);
    lcd_strobe();
    __delay_ms(10);
    lcd_strobe();
    __delay_ms(10);
    
    lcd_port &= 0x0f;   // Clear upper bits
    lcd_port |= 0x30;   // Send data to D7-D4
    lcd_strobe();
    __delay_ms(50);
    
    lcd_cmd(0x28);      // 4-bit, 2-line, 5x7 font
    lcd_cmd(0x0e);      // Display on, cursor on, blink off
    lcd_cmd(0x06);      // Autoincrement, no shift
    lcd_cmd(0x80);      // address DDRAM with 0 offset 80h
}

// Function to write a string to the LCD
void lcd_writestr(const unsigned char *c)
{
    while (*c != '\0')
    {
        lcd_data(*c);
        c++;
    }
}

// Function to place the cursor on the LCD
void lcd_setcursor(unsigned char loc)
{
    lcd_cmd(loc | 0x80);
}

// Function to move the cursor to 1st place on line 1
void lcd_line1(void)
{
    lcd_setcursor(lcdline1);
}

// Function to place cursor on 1st place on line 2
void lcd_line2(void)
{
    lcd_setcursor(lcdline2);
}

// Show awesome welcomemessage
void showWelcome(void)
{
    lcd_line1();
    lcd_writestr("  PIC 16F628A");
    lcd_line2();
    lcd_writestr("LET THE AIR FLOW"); 
    __delay_ms(2000);
    lcd_line2();
    lcd_writestr("                ");
}

// FUNCTION TO INIT THE UART PORT
void uart_init(void)
{
    TXSTAbits.BRGH = 0; // high baud selection bit, 1=high, 0=low
    TXSTAbits.SYNC = 0; // USART mode selection bit, 1=sync mode, 0=async mode
    TXSTAbits.TX9 = 0;  // 9-bit selection bit, 1=9-bit transmission, 0=8-bit transmission
    RCSTAbits.CREN = 1; // Continous receive enable bit, 1=enable continous receive
    /*
     Calculate the SPBRG with 16MHz crystal
     16MHz
     16000000 /9600 = 1666.6666
     1666.6666 / 64 = 26.0416
     26.0416 - 1= 25.0416
     25.041 = 25
     */
    
    SPBRG = 25; // 9600-n-8-1
    
    PIE1bits.RCIE = 1;  // USART receive interrupt enable bit, 1=enable
    RCSTAbits.SPEN = 1; // Serial port enable bit, 1=serial port enable
    TXSTAbits.TXEN = 1; // transmit enable bit, 1=transmit eanble
    return;
}

// Function to read a char from the uart port
unsigned char getch()
{
    getchar_active = true;
    while (getchar_active)
        continue;
    return RCREG;
}

// Interrupt function, get called each time a char is received
void interrupt ISR(void)
{
    if (RCIF && RCIE)
    {
        getchar_active = false;
        RCREG;
    }
    return;
}

// Function called when the GPS fix is not valid
void gps_waitforfix(void)
{
    lcd_line1();
    lcd_writestr("Wait for GPS    ");
    lcd_line2();
    lcd_writestr("                ");
    __delay_ms(2000);
}

// Function that turn fan on
void fan_on(void)
{
    Bat = 0;
    __delay_ms(100);
    Fan = 1;
    lcd_line2();
    lcd_writestr("   Fan is on.   ");
}

//// Function that turn fan off
//void fan_off(void)
//{
//    Fan = 0;
//    __delay_ms(1000);
//    Bat = 1;
//    lcd_line2();
//    lcd_writestr(" Batt charging. ");
//};

// MAIN PROGRAM
void main(void) {
    CMCON = 0x07;       // Disable comparators
    
    TRISA = 0b00000000; // Setting port directions
    TRISB = 0b00000110;
    PORTA = 0b00000000; // Setting all pins to LOW
    PORTB = 0b00000000;
    
    INTCONbits.PEIE = 1;    // Enable peripheral interrupts
    INTCONbits.GIE = 1;     // Enable global interrupts
    
    lcd_init();         // Initialize the LCD display
    uart_init();        // Initialize the uart port
    
    Fan = 0;            // Setting FAn pin to LOW
    Bat= 1;             // Setting the Batt pin to HIGH
    
    showWelcome();      // Show awesome welcome message
    
    while(1)            // Start program main loop - will run forever
    {
        ch = getch();   // Read the UART port, assign a char to ch
        if (ch == '$')  // Check if we have a $ sign
        {
            for (i=0;i<6;i++) // Read the first 6 chars after $
            {
                ch = getch();   // Read chars
                buffer[i] = ch; // Put the chars in a buffer
            }
            if (strcmp(buffer, gps_header)==0)  // Check if buffer is equal to out header
            {
                for (i=0;i<6;i++) // Read GPS time
                {
                    ch = getch();
                    gps_time[i] = ch;
                }
                // Check tens place and ones place of minutes 
                if (gps_time[2] == '5' && gps_time[3] >= '5')         
                {
                    fan_on(); // Turns charger off, fan on
                }
                else
                {                    
                    Fan = 0;
                    Bat = 1;
                    lcd_line2();
                    lcd_writestr(" Batt charging. ");
                }
                for (i=0;i<5;i++)
                {
                    ch = getch();
                }
                for (i=0;i<2;i++)   // Read GPS fix validation
                {
                    ch = getch();
                    gps_fix[i] = ch;
                }
                if (gps_fix[0] == 'V')
                {
                    gps_waitforfix();
                }
                lcd_line1();
                lcd_writestr("Time: ");
                lcd_data(gps_time[0]);
                lcd_data(gps_time[1]);
                lcd_data(':');
                lcd_data(gps_time[2]);
                lcd_data(gps_time[3]);
                lcd_data(':');
                lcd_data(gps_time[4]);
                lcd_data(gps_time[5]);
                lcd_writestr(" Z");                    
            }   // end strcmp
        }   // end $        
    }   // end while    
    return;
}
